
#version 150
#extension GL_ARB_separate_shader_objects : enable
#extension GL_EXT_geometry_shader4 : enable

//<vertex>

uniform mat4 matViewing;		// the viewing matrix: world space to view space

layout(location = 0) in vec4 inVertex;
out vec3 vecTexCoord;			// 3d texture coordinates to be calculated in the vertex stage and 

uniform float fStartX;			// the minimum coordinates of the bounding box of the volume in world coordinates
uniform float fStartY;
uniform float fStartZ;

uniform float fRangeX;			// the extent of the bounding box of the volume in world coordinates
uniform float fRangeY;
uniform float fRangeZ;

  // the vertex shader, which is executed once for each vertex = glyph
  void main()
  {
	  gl_Position = matViewing * inVertex;		// Transform world to viewing coordinates
	  
	  // TODO 4: Calculate texture coordinates
	  // Calculate 3D-Texture Coordinate for this glyph using the world position inVertex and the fStartXYZ and fRangeXYZ parameters.
	  // The texture coordinates should range from 0.0 to 1.0
	  
	   vecTexCoord = vec3(0.0, 0.0, 0.0); // replace this line with your calculation of the 3d texture coordinates

	  // END TODO 4
  }
  
//</vertex>


//<geometry>

layout(points) in;
layout(triangle_strip, max_vertices = 4) out;

uniform float fGlyphThreshold;			// the intensity threshold, above which a glyph will be displayed
uniform sampler3D samVolume;			// the 3d volume dataset

uniform mat4 matViewing;				// the viewing matrix: world space to view space
uniform mat4 matProjection;				// the projection matrix: view space to clip space
uniform mat4 matInverseViewing;			// the inverse viewing matrix: view space to world space
uniform float fExtentInViewspace;		// size of the billboard in view space coordinates				

uniform float fGradientScale;			// scale factor for gradient visualisation (elongation multiplier)

in vec3 vecTexCoord[];			// the 3d texture coordinates of the glyph that were calculated in the vertex stage

out float fVolumeIntensity;		// the volume intensity at the given position, between 0 and 1
out float fMagnitude;			// the gradient magnitude at the given position
out vec3 vecGradient;			// the gradient direction (normalized) at the given position

out vec4 vecWorldCoord;			// world space position of each billboard vertex, will then later be interpolated for each fragment
out vec3 vecTexCoordFrag;		  // texture coordinate of each billboard vertex, is the same for all vertices since the volume is sampled exactly once: at the glyph center point position
out vec4 vecWorldCoordBboxCenter; // world space position of the glyph bounding box center = the world space position of the billboard center point

out mat3 matRotation;			// for rotating the glyph such that the elongated axis is aligned with the gradient
out mat3 matInvRotation;		// for rotating calculation results back 

out float xs;				// The radius of the Superellipsoid in x-direction, responsible for the elongation
out float exp;				// The exponent of the Superellipsoid, responsible for the shape
out vec3 vecColor;			// The color of the Superellipsoid

mat3 getRotationMatrix(vec3 axis, float alpha); // this function will be used to set up a rotation matrix for a rotation around an arbitrary axis

// This is the geometry-program which is in our case executed once per incoming vertex (glyph), but creates a billboard consisting of four vertices
 void main()
 { 
	// TODO 5: Sample the volume
	// Sample the 3D-Volume texture at the glyph center position to look up the gradient vector (in the rgb channels) and the volume intensity (in the alpha channel).
	// Use the built-in function texture3D() to sample the volume and store the four-dimensional result in vec4 vecSample.
	// Note that the texture coordinates you calculated in the vertex shader are stored in vecTexCoord[0],
	// since this is a geometry shader with possibly many input vertices, the texture coordinates for the first (and only) vertex have to be adressed using [0]
	
	vec4 vecSample = vec4(0.01, 0.0, 0.0, 1.0); // replace this line with the texture3D-lookup at the position vecTexCoord[0]
	
	// END TODO 5
	
	
	fVolumeIntensity = vecSample.a; // the volume intensity is stored in the alpha channel, store it in an extra variable for later
	
	fMagnitude = length(vecSample.rgb);	// calculate the gradient magnitude and store it in an extra variable
	vecGradient = normalize(vecSample.rgb); // normalize the gradient direction and store it in an extra variable
	
	// TODO 6: Glyph Threshold
	// Only display a glyph at this position, if the volume intensity is high enough, otherwise terminate the pipeline for this glyph by using "return;".
	// The threshold you should use for this test is stored in the uniform variable fGlyphThreshold (between 0.0 and 1.0)
	// If this is implemented correctly, more and more glpyhs should vanish when you increase the "Glyph Threshold" in the GUI.
	
	 //if(...)   // Is the intensity value at this point of the volume large enough to become a glyph? otherwise: return

	// END TODO 6
	
	
	// TODO 7: Visual Mapping - Part 1
	// The gradient magnitude should be mapped to the major axis - radius (xs) of the ellipsoid to make the glyph elongated
	// If the gradient magnitude is 0.0, the radius should be 1.0 (isotropic shape, since the other two radii are also 1.0)
	// The larger the gradient magnitude, the larger the major axis radius (xs) should become and the more the ellipsoid is stretched
	// Additionally, multiply the gradient magnitude with the factor "fGradientScale" to visually magnify the gradient magnitude visualisation
	// If this is implemented correctly, the ellipsoids should stretch at positions where the gradient is large
	
	xs = 1.0; // replace this line by the mapping described above. Mathematically, xs is a function of gradient magnitude and fGradientScale
	 
	// END TODO 7.


	// TODO 8: Visual Mapping - Part 2
	// Additionally to the previous mapping, large gradients should become even more salient by mapping the gradient magnitude also to the shape of the glyph.
	// The shape is controlled by the exponent in the Superellipsoid equation (compare http://en.wikipedia.org/wiki/Super_ellipse) 
	// If the gradient magnitude is zero, the shape should be round, the exponent "exp" being 2.0 (like in a sphere equation)
	// The larger the gradient magnitude, the more spiky and pointed the glpyh should become. This corresponds to decreasing the exponent.
	// We only want to obtain convex shapes, thus the lower bound of the resulting exponent is 1.0. Use the max() function to achieve this.
	// Again, multiply the factor "fGradientScale" to the gradient magnitude to amplify the visualisation. 
	// If this is implemented correctly, the glyphs should become spiky at the positions where the gradient is large and round where it is small.

	exp = 2.0; // replace this line by the mapping described above.

	// END TODO 8.
	
	
	// TODO 9: Visual Mapping - Part 3
	// Map the volume intensity to the color of the glyph
	// Minimum intensity (0.0) should become fully saturated red (1,0,0)
	// Maximum intensity (1.0) should become fully saturated green (0,1,0)
	// Intensities in between should be mapped linearly to hues inbetween (green via yellow (0.5) to red)
	
	vecColor = vec3(1.0, 0.0, 0.0); // replace this line by the color mapping described above
	
	// END TODO 9.
	
	
	// In our glyph representation, the glpyh will be stretched in x-direction and then rotated such that the elongated axis aligns with the gradient
	// In fact, not the glyph is rotated but much rather everything else is (view direction, light direction), which yields the same result but allows simpler calculation
	
	// Calculate the rotation matrix, that accounts for aligning the glyph with the gradient:
	
	vec3 rotAxis = normalize(vec3(0.0, -vecGradient.z, vecGradient.y));  // this rotation axis is orthogonal to the x-axis
	float alpha = acos(vecGradient.x); // formula for angle between x axis and gradient ( cos alpha = a*b/|a|*|b|) with a = (1,0,0)	
	if(abs(alpha)>0.0001)
	{
		matRotation = getRotationMatrix(rotAxis, -alpha); // construct rotation matrix, negative angle because we don't rotate the glyph but everything else
		matInvRotation = getRotationMatrix(rotAxis, alpha); // construct inverse rotation matrix for rotating results like the intersection point back
	}
	else
	{
		matRotation = mat3(1.0); // identity, if no rotation necessary, since gradient points in x-direction anyway
		matInvRotation = matRotation;
	}
	
	// The glyph will be rendered by raycasting in the fragment shader, so we need billboards to become fragments later:
	vec4 disp[4]; // billboard xy-displacement vectors, adding them to the point position will yield the billboard vertices
	
	disp[0] = vec4(-fExtentInViewspace, -fExtentInViewspace, 0.0, 0.0) * (1.0 + fMagnitude*fGradientScale); // the billboards need to be large enough to fit the elongated glyph, therefore the scaling
	disp[1] = vec4(-fExtentInViewspace, fExtentInViewspace, 0.0, 0.0) * (1.0 + fMagnitude*fGradientScale);
	disp[2] = vec4(fExtentInViewspace, -fExtentInViewspace, 0.0, 0.0) * (1.0 + fMagnitude*fGradientScale);
	disp[3] = vec4(fExtentInViewspace, fExtentInViewspace, 0.0, 0.0) * (1.0 + fMagnitude*fGradientScale);
	
	// CREATE FOUR VERTICES FOR BILLBOARD:
	for(int i=0; i<4; i++)
	{
		gl_Position.xyz = gl_in[0].gl_Position.xyz + disp[i].xyz;	// displace this vertex to the billboard corner position in view space
		gl_Position.w = 1.0;										// set homogenous coordinate to 1
		vecWorldCoord = matInverseViewing*gl_Position;				// get world coordinates of the new billboard corner vertex and store as varying variable
		gl_Position = matProjection*gl_Position;					// project corner vertex from view to clip space
		gl_Position/=gl_Position.w;									// perspective division
		vecTexCoordFrag = vecTexCoord[0];								// pass on texture coordinates..
		vecWorldCoordBboxCenter = matInverseViewing*gl_in[0].gl_Position; // world coordinate of bounding box center = point center
		EmitVertex();													//Send finished vertex into pipeline for fragment program
	}

    EndPrimitive(); // Now we created a billboard quad (represented as triangle strip) that will allow raycasting in the fragment shader
 }

// Calculates rotation matrix for rotating about arbitrary axis and angle
 mat3 getRotationMatrix(vec3 axis, float alpha)
 {
	float cosA = cos(alpha);
	float oneMinusCosA = 1.0 - cosA;
	float sinA = sin(alpha);
 
	mat3 rotMat = 
	mat3(
		cosA + axis.x*axis.x*oneMinusCosA,        axis.x*axis.y*oneMinusCosA - axis.z*sinA,   axis.x*axis.z*oneMinusCosA + axis.y*sinA,
		axis.y*axis.x*oneMinusCosA + axis.z*sinA, cosA + axis.y*axis.y*oneMinusCosA,          axis.y*axis.z*oneMinusCosA - axis.x*sinA,
		axis.z*axis.x*oneMinusCosA - axis.y*sinA, axis.z*axis.y*oneMinusCosA + axis.x*sinA,   cosA + axis.z*axis.z*oneMinusCosA  	
	);
	
	return rotMat;
 }

//</geometry>


//<fragment>

uniform sampler3D samVolume;				// the volume dataset
uniform float fExtentInWorldspace;			// the billboard size, needed for deriving glyph bounding box dimensions
uniform vec3 vecCameraPosition;				// the camera position in world space
uniform mat4 matViewing;					// the viewing matrix: world space to view space
uniform mat4 matProjection;					// the projection matrix: view space to clip space

in float fVolumeIntensity;				// the intensity value of the volume at the glyph position
in float fMagnitude;					// the gradient magnitude at the glyph position
in vec3 vecGradient;					// the gradient direction (normalized) at the glyph position

in vec4 vecWorldCoord;					// the world space position of this fragment
in vec3 vecTexCoordFrag;				// the texture coordinates of this fragment
in vec4 vecWorldCoordBboxCenter;		// the world space position of the glyph center = bounding box center

in mat3 matRotation;					// the matrix to rotate the glyph according to the gradient direction (or rather: rotate the ray and light around the glyph)
in mat3 matInvRotation;				// for rotation calculation results back

in float xs;							// The radius of the Superellipsoid in x-direction, responsible for the elongation
in float exp;							// The exponent of the Superellipsoid, responsible for the shape
in vec3 vecColor;						// The color of the Superellipsoid

uniform mat4 matNormalInverse;				// The inverse normal matrix = transposed viewing matrix which transforms vectors from view to world space

uniform float fShadingIntensity;			// phong illumination parameters
uniform float fDiffuseShading;
uniform float fSpecularShading;
uniform float fSpecularExponent;
uniform vec3 vecLight;						// the light direction in view space as obtained from the GUI: e.g. (0,0,1) means light position = eye position

uniform float fGradientScale;				// a scale factor that determines how strongly the gradient should be emphasized by stretching the glyph

// Function prototypes:
float intersectRayBox(vec3 rayOrigin, vec3 rayDirection, vec3 box_max, vec3 box_min);
float intersectRaySuperellipsoid(vec3 vecRayOrigin, vec3 vecRayDirection, float xs, float exp);
vec3 getSuperellipsoidNormalVector(vec3 vecIntersectionPoint, float xs, float exp);
vec3 shade(vec3 vecColor, vec3 vecNormal, vec3 vecView);

// This is the fragment program that is executed once per fragment, performs raycasting to render glyph and obtains normal for shading
	void main()
	{
		// The raycasting of the glyphs will be done in this fragment program. 
		
		// The ellipsoid needs to be rotated to align with the gradient direction.
		// We do not actually rotate the Ellipsoid but rather the ray which is intersected with it, by rotating its start- and endpoint around the glyph (bbox) center
		vec3 vecRotatedCameraPosition = matRotation*(vecCameraPosition.xyz - vecWorldCoordBboxCenter.xyz) + vecWorldCoordBboxCenter.xyz;
		vec3 vecRotatedWorldCoord = matRotation*(vecWorldCoord.xyz - vecWorldCoordBboxCenter.xyz) + vecWorldCoordBboxCenter.xyz;
		
		vec3 vecRayDirection = normalize(vecRotatedWorldCoord - vecRotatedCameraPosition); // this is in world space
		
		// This is the intersection of the ray with the Superellipsoid, the return value is the parameter of the (parameter-form of the) ray where it intersects the Superellipsoid
		float fRayParam = intersectRaySuperellipsoid(vecRotatedCameraPosition.xyz, vecRayDirection.xyz, xs, exp);
				
		// check if there was an intersection:
		if(fRayParam>0.0)
		{
			vec3 vecRotatedIntersectionPoint = vecRotatedCameraPosition + fRayParam*vecRayDirection; // intersection point in world coordinates, but with rotated view ray
		
			vec3 vecNormal = getSuperellipsoidNormalVector(vecRotatedIntersectionPoint.xyz, xs, exp); // Normal in world coordinates
			
			gl_FragColor.rgb = shade(vecColor, vecNormal, -vecRayDirection); // Shading
			gl_FragColor.a = 1.0;	// Alpha value set to 100%, no blending
			
			// now the depth has to be written for z-testing. First, rotate the intersection point to obtain its correct world coordinates, as viewn in the real camera
			vec3 vecIntersectionPoint = matInvRotation*(vecRotatedIntersectionPoint.xyz - vecWorldCoordBboxCenter.xyz) + vecWorldCoordBboxCenter.xyz;
			vec4 vecProjected = (matProjection*matViewing)*vec4(vecIntersectionPoint.xyz,1.0); // project it
			vecProjected.xyz /= vecProjected.w; // perspective division
			gl_FragDepth = (1.0+vecProjected.z)*0.5; // to the [0,1] domain
			
		}
		else // no intersection with the superellipsoid
		{
			discard; // no fragment is displayed here
		}
		
	}
	
	vec3 shade(vec3 vecColor, vec3 vecNormal, vec3 vecView)
	{
		vec3 vecShaded = vecColor; // initialize shading with color

		// transform light direction (which comes from the GUI in viewspace) to worldspace:
		vec3 vecLightDirectionInWorldspace = (matNormalInverse * vec4(vecLight, 0.0)).xyz;

		// Rotate light direction, because the viewing ray was rotated before instead of the Superellipsoid. (i.e. Superellipsoid remains static, everyhting else is rotated)
		vec3 vecRotatedLightDirection = normalize(matRotation * vecLightDirectionInWorldspace);

		float fDiffuse = max(0.0,dot(vecRotatedLightDirection,vecNormal)); // diffuse shading (lambert)
		vecShaded.rgb = mix(vecColor.rgb,vecColor.rgb*fDiffuse, fDiffuseShading);
		
		vec3 vecReflection = reflect(-vecRotatedLightDirection, vecNormal); // specular reflection direction of the light in world space
		
		float fSpecular = pow(max(0.0,dot(vecReflection,vecView)),fSpecularExponent); // specular shading (phong)
		vecShaded.rgb += fSpecular*fSpecularShading;
		
		vecShaded.rgb = mix(vecColor.rgb,vecShaded.rgb, fShadingIntensity); // Apply the shading intensity modulation from the GUI
		
		return vecShaded;	
	}
	
	// This returns the parameter of the paramter-form of the ray at which the bounding box of the glyph is intersected
	float intersectRayBox(vec3 vecRayOrigin, vec3 vecRayDirection, vec3 box_max, vec3 box_min)
	{   
		vec3 tmin = (box_min - vecRayOrigin)/vecRayDirection;   
		vec3 tmax = (box_max - vecRayOrigin)/vecRayDirection;      

		vec3 real_min = min(tmin, tmax);   
		vec3 real_max = max(tmin, tmax);      
		float minmax = min( min(real_max.x, real_max.y), real_max.z);   
		float maxmin = max( max(real_min.x, real_min.y), real_min.z);   
		if (minmax >= maxmin)
			return maxmin;
		else 
			return 0.0; // no intersection found
	}
	
	// This returns the scalar value obtained by evaluating the given Superellipsoid equation at the given Position
	float evaluateSuperellipsoid(vec3 vecPos, vec3 vecCenter, float xs, float exp)
	{
		// TODO 3: Evaluating the Superellipsoid equation
		// Our special case of the Superellipsoid equation is: |(x-Cx)/xs|^exp + |y-Cy|^exp + |z-Cz|^exp = radius^exp,
		// where Cxyz is the center position of the Superellipsoid.
		// This function should evaluate the left side of this equation and return the resulting scalar float value.
		// The meaning of this value is: 
		//   left side < radius^exp --> point is inside the ellipsoid 
		//   left side == radius^exp --> point lies on the surface of the ellipsoid
		//   left side > radius^exp --> point lies outside of the ellipsoid
		
		// You should now implement the mathematical evaluation of the left side in the equation:

		 return pow(xs, exp); // replace this line by the mathematical evaluation of the left side of the Superellipsoid equation as described above

		// END TODO 3.
	} 
	
	// This intersects the (transformed) viewing ray with the Superellispoid and returns the ray-parameter of the intersection point
	float intersectRaySuperellipsoid(vec3 vecRayOrigin, vec3 vecRayDirection, float xs, float exp)
	{
		float r = pow(fExtentInWorldspace, exp); // "r" is the right side of the Superellipsoid equation, which is normally just 1.0^exp, but since we want a specific size, it's a size constant to the power of "exp"
		float errorBound = pow(1.01*fExtentInWorldspace, exp) - r; // this is a 1% approximation error bound, that, if deceeded, terminates the iteration
		
		// First, the bounding box of the Superellipsoid will be intersected as an initial solution:
		vec3 vBoxMax = vecWorldCoordBboxCenter.xyz + vec3(fExtentInWorldspace) * (1.0 + fMagnitude*fGradientScale);
		vec3 vBoxMin = vecWorldCoordBboxCenter.xyz - vec3(fExtentInWorldspace) * (1.0 + fMagnitude*fGradientScale);

		float s = intersectRayBox(vecRayOrigin.xyz, vecRayDirection.xyz, vBoxMax, vBoxMin); // "s" is the parameter of the ray where it hits the bbox
		
		// This is the vector pointing from the glyph center to the ray origin, which is needed for the following special case check:
		vec3 vecAdj = vecRayOrigin - vecWorldCoordBboxCenter.xyz;
		
		// special case - ray origin within bounding box, then return no intersection and don't render anything here
		if( abs(vecAdj.x)<xs*fExtentInWorldspace && abs(vecAdj.y)<fExtentInWorldspace && abs(vecAdj.z)<fExtentInWorldspace )  
			return 0.0;
		
		// Now the intersection of the ray with the superellipsoid will be calculated iteratively:
			
		// initial solution of secant method by evaluating the Superellipsoid at the (bbox-)intersection point:	

		vec3 vecCandidateIntersectionPoint = vecRayOrigin + vecRayDirection*s;
		float candidateDistance = evaluateSuperellipsoid(vecCandidateIntersectionPoint.xyz, vecWorldCoordBboxCenter.xyz, xs, exp);
							
							
		// The error resulting from the candidate solution:
		float resultError = candidateDistance - r;
			
		if(resultError <= errorBound) // we're good enough already, then terminate:
			return s;
			
		// iterative solution:
		int numIterations = 0;
		float sOld = s;
		s += 0.001; // increase ray parameter a little and start iteration:
		
		while(numIterations < 50)
		{
			vecCandidateIntersectionPoint = vecRayOrigin + vecRayDirection*s;
			candidateDistance = evaluateSuperellipsoid(vecCandidateIntersectionPoint.xyz, vecWorldCoordBboxCenter.xyz, xs, exp);
			
			float oldResultError = resultError;
			resultError = candidateDistance - r;
							
			if(resultError >= oldResultError) // Superellipsoid will not be hit, iteration diverges, return 0
				return 0.0;
			
			if(resultError < errorBound) // We are good enough! Terminate loop and return current ray parameter "s".
				break;
				
			float correction = (resultError * ( s - sOld )) / ( resultError - oldResultError ); // this correction factor is the core of the secant method
			sOld = s;
			
			s -= correction;
		
			numIterations++;
		}	
		return s; // finally return the current solution as approximation of the intersection point
	
	}
	
	// This approximates the normal vector of the Superellipsoid surface at the given point
	vec3 getSuperellipsoidNormalVector( vec3 vecIntersectionPoint, float xs, float exp)
	{
		vec3 vecNormal;
	
		vecNormal.x = ( vecIntersectionPoint.x - vecWorldCoordBboxCenter.x );
		vecNormal.y = ( vecIntersectionPoint.y - vecWorldCoordBboxCenter.y ) * xs;
		vecNormal.z = ( vecIntersectionPoint.z - vecWorldCoordBboxCenter.z ) * xs;
		
		if( vecNormal.x > 0.0 ) 
			vecNormal.x = pow( vecNormal.x, exp - 1.0 );
		else 
			vecNormal.x = -pow( -vecNormal.x, exp - 1.0 );
			
		if( vecNormal.y > 0.0 ) 
			vecNormal.y = pow( vecNormal.y, exp - 1.0 );
		else 
			vecNormal.y = -pow( -vecNormal.y, exp - 1.0 );
		
		if( vecNormal.z > 0.0 ) 
			vecNormal.z = pow( vecNormal.z, exp - 1.0 );
		else 
			vecNormal.z = -pow( -vecNormal.z, exp - 1.0 );
		
		return normalize(vecNormal);
	}

//</fragment>
